from mygrad.tensor_base import Tensor

from .ops import *

__all__ = [
    "exp",
    "exp2",
    "expm1",
    "logaddexp",
    "logaddexp2",
    "log",
    "log2",
    "log10",
    "log1p",
]


def exp(a, constant=False):
    """ ``f(a) -> exp(a)``

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Exp, a, constant=constant)


def exp2(a, constant=False):
    """``f(a) -> 2^a``

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)
        
        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Exp2, a, constant=constant)


def expm1(a, constant=False):
    """ ``f(a) -> exp(a) - 1``

        The inverse of ``logp1``.

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=True)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Notes
        -----
        This function provides greater precision than ``exp(x) - 1`` for
        small values of ``x``."""
    return Tensor._op(Expm1, a, constant=constant)


def logaddexp(a, b, constant=False):
    """ ``f(a, b) -> log(exp(a) + exp(b))``

        Parameters
        ----------
        a : array_like

        b : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Notes
        -----
        This function is useful
        in statistics where the calculated probabilities of events may
        be so small as to exceed the range of normal floating point
        numbers. In such cases the logarithm of the calculated
        probability is stored. This function allows adding probabilities
        stored in such a fashion. """
    return Tensor._op(Logaddexp, a, b, constant=constant)


def logaddexp2(a, b, constant=False):
    """ ``f(a, b) -> log_2(2 ** a + 2 ** b)``

        Utilizes base-2 log.

        Parameters
        ----------
        a : array_like

        b : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Notes
        -----
        This function is useful
        in statistics where the calculated probabilities of events may
        be so small as to exceed the range of normal floating point
        numbers. In such cases the logarithm of the calculated
        probability is stored. This function allows adding probabilities
        stored in such a fashion. """
    return Tensor._op(Logaddexp2, a, b, constant=constant)


def log(a, constant=False):
    """ ``f(a) -> log(a)``

        Natural logarithm

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=True)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Notes
        -----
        This function provides greater precision than ``exp(x) - 1`` for
        small values of ``x``."""
    return Tensor._op(Log, a, constant=constant)


def log2(a, constant=False):
    """ ``f(a) -> log2(a)``

        Base-2 logarithm

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=True)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Log2, a, constant=constant)


def log10(a, constant=False):
    """ ``f(a) -> log10(a)``

        Base-10 logarithm

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=True)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Log10, a, constant=constant)


def log1p(a, constant=False):
    """ f(a) -> log(1 + a)

        The inverse of ``expm1``.

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=True)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Notes
        -----
        For real-valued input, log1p is accurate also
        for ``x`` so small that ``1 + x == 1`` in floating-point
        accuracy."""
    return Tensor._op(Log1p, a, constant=constant)



from mygrad.tensor_base import Tensor

from .ops import *

__all__ = ["reshape", "squeeze", "ravel", "expand_dims", "broadcast_to"]


def reshape(a, *newshape, constant=False):
    """ Returns a tensor with a new shape, without changing its data.

        This docstring was adapted from ``numpy.reshape``

        Parameters
        ----------
        a : array_like
            The tensor to be reshaped

        *newshape : Union[int, Tuple[int, ...]]
            The new shape should be compatible with the original shape. If
            an integer, then the result will be a 1-D tensor of that length.
            One shape dimension can be -1. In this case, the value is
            inferred from the length of the tensor and remaining dimensions.

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor
            ``a`` with its shape changed permuted.  A new tensor is returned.

        Notes
        -----
        ``reshape`` utilizes C-ordering, meaning that it reads & writes elements using
        C-like index ordering; the last axis index changing fastest, and, proceeding
        in reverse order, the first axis index changing slowest. 
            
        Examples
        --------
        >>> import mygrad as mg
        >>> a = mg.Tensor([[1,2,3], [4,5,6]])
        >>> mg.reshape(a, 6)
        Tensor([1, 2, 3, 4, 5, 6])

        >>> mg.reshape(a, (3,-1))   # the unspecified value is inferred to be 2
        Tensor([[1, 2],
                [3, 4],
                [5, 6]])"""
    if not newshape:
        raise TypeError("reshape() takes at least 1 argument (0 given)")
    if hasattr(newshape[0], "__iter__"):
        if len(newshape) > 1:
            raise TypeError("an integer is required")
        newshape = newshape[0]
    return Tensor._op(Reshape, a, op_args=(newshape,), constant=constant)


def squeeze(a, axis=None, constant=False):
    """
    Remove single-dimensional entries from the shape of a tensor.

    This docstring was adapted from ``numpy.squeeze``

    Parameters
    ----------
    a : array_like
        The tensor to be reshaped
    
    axis : Optional[int, Tuple[int, ...]]
        Selects a subset of the single-dimensional entries in the 
        shape. If an axis is selected with shape entry greater than 
        one, an error is raised.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Raises
    ------
    ValueError
        If ``axis`` is not ``None``, and an axis being squeezed is not of length 1

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.Tensor([[[0], [1], [2]]])
    >>> x.shape
    (1, 3, 1)
    >>> mg.squeeze(x).shape
    (3,)
    >>> mg.squeeze(x, axis=0).shape
    (3, 1)
    >>> mg.squeeze(x, axis=1).shape
    Traceback (most recent call last):
    ...
    ValueError: cannot select an axis to squeeze out which has size not equal to one
    >>> mg.squeeze(x, axis=2).shape
    (1, 3)"""
    return Tensor._op(Squeeze, a, op_args=(axis,), constant=constant)


def ravel(a, constant=False):
    """
    Flattens contents of a tensor into a contiguous 1-D array.  A copy is made only if needed.

    This docstring was adapted from ``numpy.ravel``.

    Parameters
    ----------
    a : array_like
        The tensor to be flattened

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Notes
    -----
    ``ravel`` utilizes C-ordering, meaning that it reads & writes elements using
    C-like index ordering; the last axis index changing fastest, and, proceeding
    in reverse order, the first axis index changing slowest.

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.Tensor([[1, 2],
    ...                [3, 4]])
    >>> mg.ravel(x)
    Tensor([1, 2, 3, 4])
    """
    return Tensor._op(Ravel, a, constant=constant)


def expand_dims(a, axis, constant=False):
    """
    Expand the dimensions of a tensor by adding a new axis.

    This docstring was adapted from ``numpy.expand_dims``.

    Parameters
    ----------
    a : array_like
        The tensor to be expanded

    axis : int
        The position of the new axis in the expanded array shape.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.Tensor([1, 2])
    >>> x.shape
    (2,)
    >>> y = mg.expand_dims(x, 1)
    >>> y.shape
    (2, 1)
    >>> z = mg.expand_dims(y, 0)
    >>> z.shape
    (1, 2, 1)
    """
    return Tensor._op(ExpandDims, a, op_args=(axis,), constant=constant)


def broadcast_to(a, shape, constant=False):
    """
    Broadcast a tensor to a new shape.

    This docstring was adapted from ``numpy.broadcast_to``.

    Parameters
    ----------
    a : array_like
        The tensor to be broadcasted

    shape: Tuple[int, ...]
        The shape of the broadcasted tensor. This shape
        should be broadcast-compatible with the original
        shape.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Raises
    ------
    ValueError
        If the array is not compatible with the new shape
        according to Numpy's broadcasting rules.

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.Tensor([1, 2, 3])
    >>> mg.broadcast_to(x, (3,3))
    Tensor([[1, 2, 3],
            [1, 2, 3],
            [1, 2, 3]])
    >>> mg.broadcast_to(x, (4,4))
    Traceback (most recent call last):
    ...
    ValueError: operands could not be broadcast together with remapped
    shapes [original->remapped]: (3,) and requested shape (4,4)
    """
    return Tensor._op(BroadcastTo, a, op_args=(shape,), constant=constant)



import numpy as np
from numpy.core.einsumfunc import _parse_einsum_input

import mygrad as mg
from mygrad import Tensor

from .ops import *

__all__ = ["multi_matmul", "matmul", "einsum"]


def matmul(a, b, constant=False):
    r"""
    Matrix product of two tensors:

    ``matmul(x, y)`` is equivalent to ``x @ y``.

    This documentation was adapted from ``numpy.matmul``

    The behavior depends on the arguments in the following way.

    - If both arguments are 2-D they are multiplied like conventional
      matrices.
    - If either argument is N-D, N > 2, it is treated as a stack of
      matrices residing in the last two indexes and broadcast accordingly.
    - If the first argument is 1-D, it is promoted to a matrix by
      prepending a 1 to its dimensions. After matrix multiplication
      the prepended 1 is removed.
    - If the second argument is 1-D, it is promoted to a matrix by
      appending a 1 to its dimensions. After matrix multiplication
      the appended 1 is removed.

    Multiplication by a scalar is not allowed, use ``*`` instead. Note that
    multiplying a stack of matrices with a vector will result in a stack of
    vectors, but matmul will not recognize it as such.

    ``matmul`` differs from ``numpy.dot`` in two important ways.

    - Multiplication by scalars is not allowed.
    - Stacks of matrices are broadcast together as if the matrices
      were elements.


    Parameters
    ----------
    a : array_like

    b : array_like

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    output : mygrad.Tensor
        Returns the matrix product of `a` and `b`.  If `a` and `b` are both
        1-D arrays then a scalar is returned; otherwise an array is
        returned.


    Raises
    ------
    ValueError
        If the last dimension of `a` is not the same size as
        the second-to-last dimension of `b`.

        If scalar value is passed.

    Notes
    -----
    The matmul function implements the semantics of the `@` operator introduced
    in Python 3.5 following PEP465.

    Examples
    --------
    For two 2D tensors, ``matmul(a, b)`` is the matrix product :math:`\sum_{j}{A_{ij} B_{jk}} = F_{ik}`:

    >>> import mygrad as mg
    >>> a = [[1, 0], [0, 1]]
    >>> b = [[4, 1], [2, 2]]
    >>> mg.matmul(a, b)
    Tensor([[4, 1],
            [2, 2]])

    For 2-D mixed with 1-D, the result is the matrix-vector product, :math:`\sum_{j}{A_{ij} B_{j}} = F_{i}`:

    >>> a = [[1, 0], [0, 1]]
    >>> b = [1, 2]
    >>> mg.matmul(a, b)
    Tensor([1, 2])

    Broadcasting is conventional for stacks of arrays. Here ``a`` is treated
    like a stack of three 5x6 matrices, and the 6x4 matrix ``b`` is broadcast
    matrix-multiplied against each one. This produces a shape-(3, 5, 4) tensor
    as a result.

    >>> a = mg.arange(3*5*6).reshape((3,5,6))
    >>> b = mg.arange(6*4).reshape((6,4))
    >>> mg.matmul(a,b).shape
    (3, 5, 4)

    Scalar multiplication raises an error.

    >>> mg.matmul(a, 3)
    Traceback (most recent call last):
    ...
    ValueError: Scalar operands are not allowed, use '*' instead"""
    return Tensor._op(MatMul, a, b, constant=constant)


def einsum(*operands, optimize=False, constant=False):
    r"""
    einsum(subscripts, *operands)

    Evaluates the Einstein summation convention on the operands. This implementation
    exactly mirrors that of ``numpy.einsum`` and supports back-propagation through
    all variety of tensor-products, sums, traces, and views that it can perform.

    The following docstring was adapted from the documentation for ``numpy.einsum``

    Using the Einstein summation convention, many common multi-dimensional
    array operations can be represented in a simple fashion.  This function
    provides a way to compute such summations. The best way to understand this
    function is to try the examples below, which show how many common NumPy/MyGrad
    functions can be implemented as calls to ``einsum``.

    Back-propagation via ``einsum`` is optimized such that any tensor that occurs
    redundantly within the summation will only have its gradient computed once.
    This optimization accommodates all number and combination of redundancies that can
    be encountered.

    E.g. back-propping through ``einsum('...,...->', x, x)`` will only incur a single
    computation/accumulation for ``x.grad`` rather than two. This permits users to
    leverage the efficiency of sum-reduction, where ``(x ** 2).sum()`` is sub-optimal,
    without being penalized during back-propagation.

    Parameters
    ----------
    subscripts : str
        Specifies the subscripts for summation.

    operands : array_like
        The tensors used in the summation.

    optimize : {False, True, 'greedy', 'optimal'}, optional (default=False)
        Controls if intermediate optimization should occur; also enables
        the use of BLAS where possible. This can produce significant speedups
        for computations like matrix multiplication.

        No optimization will occur if False and True will default to the 'greedy'
        algorithm. Also accepts an explicit contraction list from the
        ``np.einsum_path`` function. See ``np.einsum_path`` for more details.

    constant : bool, optional (default=False)
        If True, the resulting Tensor is a constant.

    Returns
    -------
    output : mygrad.Tensor
        The calculation based on the Einstein summation convention.

    Notes
    -----
    The subscripts string is a comma-separated list of subscript labels,
    where each label refers to a dimension of the corresponding operand.
    Repeated subscripts labels in one operand take the diagonal.  For example,
    ``einsum('ii', a)`` is equivalent to ``np.trace(a)`` (however, the former
    supports back-propagation).

    Whenever a label is repeated, it is summed, so ``einsum('i, i', a, b)``
    is equivalent to ``np.inner(a, b)``.  If a label appears only once,
    it is not summed, so ``einsum('i', a)`` produces a view of ``a``
    with no changes.

    The order of labels in the output is by default alphabetical.  This
    means that ``np.einsum('ij', a)`` doesn't affect a 2D tensor, while
    ``einsum('ji', a)`` takes its transpose.

    The output can be controlled by specifying output subscript labels
    as well.  This specifies the label order, and allows summing to
    be disallowed or forced when desired.  The call ``einsum('i->', a)``
    is like ``np.sum(a, axis=-1)``, and ``einsum('ii->i', a)``
    is like ``np.diag(a)``.  The difference is that `einsum` does not
    allow broadcasting by default.

    To enable and control broadcasting, use an ellipsis.  Default
    NumPy-style broadcasting is done by adding an ellipsis
    to the left of each term, like ``einsum('...ii->...i', a)``.
    To take the trace along the first and last axes,
    you can do ``einsum('i...i', a)``, or to do a matrix-matrix
    product with the left-most indices instead of rightmost, you can do
    ``einsum('ij...,jk...->ik...', a, b)``.

    When there is only one operand, no axes are summed, and no output
    parameter is provided, a view into the operand is returned instead
    of a new tensor.  Thus, taking the diagonal as ``einsum('ii->i', a)``
    produces a view.

    An alternative way to provide the subscripts and operands is as
    ``einsum(op0, sublist0, op1, sublist1, ..., [sublistout])``. The examples
    below have corresponding `einsum` calls with the two parameter methods.

    Examples
    --------
    >>> import mygrad as mg
    >>> import numpy as np
    >>> a = mg.arange(25).reshape(5,5)
    >>> b = mg.arange(5)
    >>> c = mg.arange(6).reshape(2,3)

    Compute the trace of ``a``, :math:`\sum_{i}{A_{ii}} = f`:

    >>> einsum('ii', a)
    Tensor(60)
    >>> einsum(a, [0, 0])
    Tensor(60)
    >>> np.trace(a.data)
    array(60)

    Return a view along the diagonal of ``a``, :math:`A_{ii} = F_{i}`:

    >>> einsum('ii->i', a)
    Tensor([ 0,  6, 12, 18, 24])
    >>> einsum(a, [0,0], [0])
    Tensor([ 0,  6, 12, 18, 24])
    >>> np.diag(a.data)
    array([ 0,  6, 12, 18, 24])

    Compute the matrix-vector product of ``a`` with ``b``, :math:`\sum_{j}{A_{ij} B_{j}} = F_{i}`:

    >>> einsum('ij,j', a, b)
    Tensor([ 30,  80, 130, 180, 230])
    >>> einsum(a, [0,1], b, [1])
    Tensor([ 30,  80, 130, 180, 230])
    >>> mg.matmul(a, b)
    Tensor([ 30,  80, 130, 180, 230])
    >>> einsum('...j,j', a, b)
    Tensor([ 30,  80, 130, 180, 230])

    Take the transpose of ``c``, :math:`C_{ji} = F_{ij}`:

    >>> einsum('ji', c)
    Tensor([[0, 3],
            [1, 4],
            [2, 5]])
    >>> einsum(c, [1, 0])
    Tensor([[0, 3],
            [1, 4],
            [2, 5]])
    >>> c.T
    Tensor([[0, 3],
            [1, 4],
            [2, 5]])

    Compute ``3 * c``:

    >>> einsum('..., ...', 3, c)
    Tensor([[ 0,  3,  6],
            [ 9, 12, 15]])
    >>> einsum(',ij', 3, c)
    Tensor([[ 0,  3,  6],
            [ 9, 12, 15]])
    >>> einsum(3, [Ellipsis], c, [Ellipsis])
    Tensor([[ 0,  3,  6],
            [ 9, 12, 15]])
    >>> 3 * c
    Tensor([[ 0,  3,  6],
            [ 9, 12, 15]])

    Compute the inner product of ``b`` with itself, :math:`\sum_{i}{B_{i} B_{i}} = f`:

    >>> einsum('i,i', b, b)
    Tensor(30)
    >>> einsum(b, [0], b, [0])
    Tensor(30)
    >>> np.inner(b.data, b.data)
    30

    Compute the outer product of ``array([1, 2])`` with ``b``, :math:`A_{i}B_{j} = F_{ij}`:

    >>> einsum('i,j', np.arange(2)+1, b)
    Tensor([[0, 1, 2, 3, 4],
           [0, 2, 4, 6, 8]])
    >>> einsum(np.arange(2)+1, [0], b, [1])
    Tensor([[0, 1, 2, 3, 4],
           [0, 2, 4, 6, 8]])
    >>> np.outer(np.arange(2)+1, b)
    array([[0, 1, 2, 3, 4],
           [0, 2, 4, 6, 8]])
    >>> einsum('i...->...', a)
    Tensor([50, 55, 60, 65, 70])
    >>> einsum(a, [0,Ellipsis], [Ellipsis])
    Tensor([50, 55, 60, 65, 70])
    >>> np.sum(a, axis=0)
    array([50, 55, 60, 65, 70])

    Compute the tensor product :math:`\sum_{ij}{A_{ijk} B_{jil}} = F_{kl}`

    >>> a = mg.arange(60.).reshape(3,4,5)
    >>> b = mg.arange(24.).reshape(4,3,2)
    >>> einsum('ijk,jil->kl', a, b)
    Tensor([[ 4400.,  4730.],
            [ 4532.,  4874.],
            [ 4664.,  5018.],
            [ 4796.,  5162.],
            [ 4928.,  5306.]])
    >>> einsum(a, [0,1,2], b, [1,0,3], [2,3])
    Tensor([[ 4400.,  4730.],
            [ 4532.,  4874.],
            [ 4664.,  5018.],
            [ 4796.,  5162.],
            [ 4928.,  5306.]])
    >>> np.tensordot(a,b, axes=([1,0],[0,1]))
    array([[ 4400.,  4730.],
            [ 4532.,  4874.],
            [ 4664.,  5018.],
            [ 4796.,  5162.],
            [ 4928.,  5306.]])

    Matrix multiply ``a.T`` with ``b.T``, :math:`\sum_{k}{A_{ki} B_{jk}} = F_{ij}`

    >>> a = mg.arange(6).reshape((3,2))
    >>> b = mg.arange(12).reshape((4,3))
    >>> einsum('ki,jk->ij', a, b)
    Tensor([[10, 28, 46, 64],
            [13, 40, 67, 94]])
    >>> einsum('ki,...k->i...', a, b)
    Tensor([[10, 28, 46, 64],
            [13, 40, 67, 94]])
    >>> einsum('k...,jk', a, b)
    Tensor([[10, 28, 46, 64],
            [13, 40, 67, 94]])

    Make an assignment to a view along the diagonal of ``a``:

    >>> a = mg.zeros((3, 3))
    >>> einsum('ii->i', a).data[:] = 1
    >>> a
    Tensor([[ 1.,  0.,  0.],
            [ 0.,  1.,  0.],
            [ 0.,  0.,  1.]])
    """

    # TODO: normalize error handling for invalid inputs
    operands = list(operands)
    if isinstance(operands[0], str):
        # operands form: "ijk, ijk", x, y
        variables = operands[1:]
        if any(isinstance(i, Tensor) for i in operands):
            operands[1:] = (
                var.data if isinstance(var, Tensor) else var for var in operands[1:]
            )
    else:
        # operands form: op0, sublist0, op1, sublist1, ..., [sublistout]
        end = -1 if len(operands) % 2 else None  # -1 if sublistout is included
        variables = operands[:end:2]
        if any(isinstance(i, Tensor) for i in operands):
            operands[:end:2] = (
                var.data if isinstance(var, Tensor) else var for var in operands[:end:2]
            )

    in_lbls, out_lbls, _ = _parse_einsum_input(operands)
    return Tensor._op(
        EinSum,
        *variables,
        op_kwargs=dict(in_lbls=in_lbls, out_lbls=out_lbls, optimize=optimize),
        constant=constant
    )


def multi_matmul(tensors, constant=False):
    """
    Matrix product of two or more tensors calculated in the optimal ordering

    Parameters
    ----------
    tensors: Sequence[array_like]
        The sequence of tensors to be matrix-multiplied.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient).

    Returns
    -------
    mygrad.Tensor
        Returns the matrix product of the tensors provided


    Extended Summary
    ----------------
    This documentation was adapted from ``numpy.linalg.multi_dot``

    Compute the matrix multiplication of two or more arrays in a single function
    call, while automatically selecting the fastest evaluation order.
    ``multi_matmul`` chains ``matmul`` and uses optimal parenthesization  [1]_ [2]_.
    Depending on the shapes of the matrices, this can speed up the multiplication a lot.

    If the first argument is 1-D it is treated as a row vector.

    If the last argument is 1-D it is treated as a column vector.

    The other arguments must be 2-D or greater.

    Think of `multi_dot` as an optimized version of::

        def multi_dot(tensors): return functools.reduce(mg.matmul, tensors)

    Raises
    ------
    ValueError
        If ``tensors`` contains less than two array_like items.

    ValueError
        If ``tensor`` other than the first or last is less than two dimensional

    See Also
    --------
    matmul : matrix multiplication with two arguments.

    References
    ----------

    .. [1] Cormen, "Introduction to Algorithms", Chapter 15.2, p. 370-378
    .. [2] http://en.wikipedia.org/wiki/Matrix_chain_multiplication

    Notes
    -----
    The cost for a matrix multiplication can be calculated with the
    following function::

        def cost(A, B):
            return A.shape[0] * A.shape[1] * B.shape[1]

    Let's assume we have three matrices :math:`A_{10x100}, B_{100x5}, C_{5x50}`.

    The costs for the two different parenthesizations are as follows::

        cost((AB)C) = 10*100*5 + 10*5*50   = 5000 + 2500   = 7500
        cost(A(BC)) = 10*100*50 + 100*5*50 = 50000 + 25000 = 75000

    Examples
    --------
    ``multi_matmul`` allows you to write:

    >>> from mygrad import multi_matmul, matmul, Tensor
    >>> import numpy as np
    >>> # Prepare some random tensors
    >>> A = Tensor(np.random.random((10000, 100)))
    >>> B = Tensor(np.random.random((100, 1000)))
    >>> C = Tensor(np.random.random((1000, 5)))
    >>> D = Tensor(np.random.random((5, 333)))
    >>> # the actual matrix multiplication
    >>> multi_matmul([A, B, C, D]) # computes (A @ (B @ C)) @ D

    instead of:

    >>> matmul(matmul(matmul(A, B), C), D)
    >>> # or
    >>> A @ B @ C @ D
    """

    for a in tensors:
        if not (1 <= a.ndim <= 2):
            raise ValueError(
                "%d-dimensional tensor given. Tensor must be one or two-dimensional"
                % (a.ndim,)
            )

    n = len(tensors)
    if n < 2:
        raise ValueError("Expecting at least two arrays.")
    elif n == 2:
        return matmul(tensors[0], tensors[1], constant)

    tensors = [a if isinstance(a, Tensor) else np.asarray(a) for a in tensors]

    # save original ndim to reshape the result array into the proper form later
    ndim_first, ndim_last = tensors[0].ndim, tensors[-1].ndim

    # Explicitly convert vectors to 2D arrays to keep the logic of this function simpler
    if tensors[0].ndim == 1:
        tensors[0] = mg.expand_dims(
            tensors[0],
            axis=0,
            constant=tensors[0].constant if isinstance(tensors[0], Tensor) else True,
        )
    if tensors[-1].ndim == 1:
        tensors[-1] = mg.expand_dims(
            tensors[-1],
            axis=1,
            constant=tensors[-1].constant if isinstance(tensors[-1], Tensor) else True,
        )

    if n == 3:
        result = _multi_matmul_three(tensors[0], tensors[1], tensors[2], constant)
    else:
        order = _multi_matmul_chain_order(tensors)
        result = _multi_matmul(tensors, order, 0, n - 1, constant)

    # return proper shape since we possibly added dimensions to the first
    # and last arrays
    if ndim_first == 1 and ndim_last == 1:
        return result[0, 0]
    elif ndim_first == 1 or ndim_last == 1:
        return result.reshape(-1)
    else:
        return result


def _multi_matmul_three(A, B, C, constant=False) -> Tensor:
    """
    Find the best order for three arrays and do the multiplication.

    """
    a0, a1b0 = A.shape[-2:]
    b1c0, c1 = C.shape[-2:]
    cost1 = a0 * b1c0 * (a1b0 + c1)
    cost2 = a1b0 * c1 * (a0 + b1c0)

    if cost1 < cost2:
        return matmul(matmul(A, B, constant), C, constant)
    else:
        return matmul(A, matmul(B, C, constant), constant)


def _multi_matmul_chain_order(arrays):
    """
    Return a np.array that encodes the optimal order of multiplications.
    The optimal order array is then used by `_multi_matmul()` to do the
    multiplication.

    The implementation CLOSELY follows Cormen, "Introduction to Algorithms",
    Chapter 15.2, p. 370-378.  Note that Cormen uses 1-based indices.

        cost[i, j] = min([
            cost[prefix] + cost[suffix] + cost_mult(prefix, suffix)
            for k in range(i, j)])
    """
    n = len(arrays)
    # p stores the dimensions of the matrices
    # Example for p: A_{10x100}, B_{100x5}, C_{5x50} --> p = [10, 100, 5, 50]
    # Using -2 to generalize for shapes that are more than 2 dimmensions
    p = [a.shape[-2] for a in arrays] + [arrays[-1].shape[-1]]
    # m is a matrix of costs of the subproblems
    # m[i,j]: min number of scalar multiplications needed to compute A_{i..j}
    m = np.zeros((n, n), dtype=np.double)
    # s is the actual ordering
    # s[i, j] is the value of k at which we split the product A_i..A_j
    s = np.empty((n, n), dtype=np.intp)

    for l in range(1, n):
        for i in range(n - l):
            j = i + l
            m[i, j] = np.inf
            for k in range(i, j):
                q = m[i, k] + m[k + 1, j] + p[i] * p[k + 1] * p[j + 1]
                if q < m[i, j]:
                    m[i, j] = q
                    s[i, j] = k  # Note that Cormen uses 1-based index
    return s


def _multi_matmul(arrays, order, i, j, constant=False) -> Tensor:
    """Actually do the multiplication with the given order."""
    if i == j:
        return arrays[i]
    else:
        return matmul(
            _multi_matmul(arrays, order, i, order[i, j], constant),
            _multi_matmul(arrays, order, order[i, j] + 1, j, constant),
            constant,
        )



from mygrad.tensor_base import Tensor

from .ops import MoveAxis, Roll, SwapAxes, Transpose

__all__ = ["transpose", "moveaxis", "swapaxes", "roll"]


def transpose(a, *axes, constant=False):
    """ Permute the dimensions of a tensor.

        Parameters
        ----------
        a : array_like
            The tensor to be transposed

        axes : Optional[Tuple[int]]
            By default, reverse the dimensions, otherwise permute the axes
            according to the values given.

        Returns
        -------
        mygrad.Tensor
            `a` with its axes permuted.  A new tensor is returned.

        Examples
        --------
        >>> import mygrad as mg
        >>> a = mg.Tensor([[1, 2], [3, 4]])
        >>> a
        Tensor([[1, 2],
                [3, 4]])
        >>> a.transpose()
        Tensor([[1, 3],
                [2, 4]])
        >>> a.transpose((1, 0))
        Tensor([[1, 3],
                [2, 4]])
        >>> a.transpose(1, 0)
        Tensor([[1, 3],
                [2, 4]]) """
    if not axes:
        axes = None
    elif hasattr(axes[0], "__iter__") or axes[0] is None:
        if len(axes) > 1:
            raise TypeError(
                "'{}' object cannot be interpreted as an integer".format(type(axes[0]))
            )
        axes = axes[0]
    return Tensor._op(Transpose, a, op_args=(axes,), constant=constant)


def moveaxis(a, source, destination, constant=False):
    """ Move axes of a tensor to new positions. Other axes remain in their
        original order.


        Parameters
        ----------
        a : array_like
            The array whose axes should be reordered.

        source : Union[int, Sequence[int]]
            Original positions of the axes to move. These must be unique.

        destination : Union[int, Sequence[int]]
            Destination positions for each of the original axes. These must also be
            unique.

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        result : mygrad.Tensor
            Array with moved axes. This array is a view of the input array..

        Examples
        --------
        >>> from mygrad import Tensor, moveaxis
        >>> x = Tensor(np.zeros((3, 4, 5)))
        >>> moveaxis(x, 0, -1).shape
        (4, 5, 3)
        >>> moveaxis(x, -1, 0).shape
        (5, 3, 4)
        >>> moveaxis(x, [0, 1], [-1, -2]).shape
        (5, 4, 3) """
    return Tensor._op(MoveAxis, a, op_args=(source, destination), constant=constant)


def swapaxes(a, axis1, axis2, constant=False):
    """ Interchange two axes of a tensor.

        Parameters
        ----------
        a : array_like
            Input array.

        axis1 : int
            First axis.

        axis2 : int
            Second axis.

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Examples
        --------
        >>> from mygrad import Tensor, swapaxes
        >>> x = Tensor([[1, 2, 3]])
        >>> swapaxes(x, 0, 1)
        Tensor([[1],
               [2],
               [3]])
        >>> x = Tensor([[[0, 1], [2, 3]], [[4, 5], [6, 7]]])
        >>> x
        Tensor([[[0, 1],
                [2, 3]],
               [[4, 5],
                [6, 7]]])
        >>> swapaxes(x, 0, 2)
        Tensor([[[0, 4],
                [2, 6]],
               [[1, 5],
                [3, 7]]])
    """
    return Tensor._op(SwapAxes, a, op_args=(axis1, axis2), constant=constant)


def roll(a, shift, axis=None, constant=False):
    """
    Roll tensor elements along a given axis.

    Elements that roll beyond the end of an axis "wrap back around" to the beginning.

    This docstring was adapted from ``numpy.roll``

    Parameters
    ----------
    a : array_like
        Input tensor.

    shift : Union[int, Tuple[int, ...]]
        The number of places by which elements are shifted.  If a tuple,
        then `axis` must be a tuple of the same size, and each of the
        given axes is shifted by the corresponding number.  If an int
        while `axis` is a tuple of ints, then the same value is used for
        all given axes.

    axis : Optional[Union[int, Tuple[int, ...]]]
        Axis or axes along which elements are shifted.  By default, the
        array is flattened before shifting, after which the original
        shape is restored.

    Returns
    -------
    res : Tensor
        Output array, with the same shape as `a`.


    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.arange(10)
    >>> mg.roll(x, 2)
    Tensor([8, 9, 0, 1, 2, 3, 4, 5, 6, 7])
    >>> x2 = mg.reshape(x, (2,5))
    >>> x2
    Tensor([[0, 1, 2, 3, 4],
           [5, 6, 7, 8, 9]])
    >>> mg.roll(x2, 1)
    Tensor([[9, 0, 1, 2, 3],
           [4, 5, 6, 7, 8]])
    >>> mg.roll(x2, 1, axis=0)
    Tensor([[5, 6, 7, 8, 9],
           [0, 1, 2, 3, 4]])
    >>> mg.roll(x2, 1, axis=1)
    Tensor([[4, 0, 1, 2, 3],
           [9, 5, 6, 7, 8]])
    """
    return Tensor._op(
        Roll, a, op_kwargs=dict(shift=shift, axis=axis), constant=constant
    )



import numpy as np
from graphviz import Digraph

from mygrad.tensor_base import Tensor


def build_graph(
    fin,
    names=None,
    *,
    render=True,
    save=False,
    dims=False,
    dtypes=False,
    sum_stats=False
):
    """ Builds and renders a computational graph.

        Parameters
        ----------
        fin : mygrad.Tensor
            The tensor object that will be the final node in the
            computational graph.

        names : Optional[Dict[str, Union[mygrad.Tensor, numpy.ndarray]]]
            A dictionary that maps names of Tensors to Tensor objects. If
            an argument is passed to names, the key name that maps to a Tensor
            included in the computational graph will be used as a label for the
            Tensor's node. If no argument is passed, the nodes on the
            computational graph will display the full Tensor.

            To use the names assigned in the local environment,
            pass ``names=locals()`` to the build_graph function.

            If different names are used from the local environment,
            the key must map to the exact Tensor object. A new Tensor or copy
            of the original Tensor should not be created as the value in the
            dictionary.

            Only instances of mygrad.Tensor or numpy.ndarray can have labels
            assigned to Nodes. If a list or tuple is used in an operation
            with a Tensor, and names is not None, the Node label will be
            set to *Constant*. If a list or tuple is used in multiple operations,
            a unique Node will be created for each time it is used.

            A scalar will always be used as the label for a 0-dimensional
            Tensor's Node.

        render : bool, optional (default=True)
            If True, build_graph will return a graphviz Digraph object that,
            when called, will render the computational graph in a Jupyter
            notebook or the Jupyter Qt console. If False, nothing is returned.

        save : bool, optional (default=False)
            If True, build_graph will save a rendered computational graph to
            the current working directory as ``computational_graph.pdf``.

        dims : bool, optional (default=False)
            If True, Tensor dimensions are added to Node labels. Dimensions
            will not be displayed for scalar values.

        dtypes : bool, optional (default=False)
            If True, Tensor data types are added to Node labels.

        sum_stats : bool, optional (default=False)
            If True, Tensor minimums, maximums, medians, and means are
            added to Node labels. These will not be displayed for scalar values.

        Returns
        -------
        Union[graphviz.Digraph, None]

        Notes
        -----
        build_graph requires that Graphviz is installed.
    """
    assert isinstance(fin, Tensor), "fin must be a Tensor"
    assert isinstance(names, (dict, type(None)))
    assert isinstance(render, bool)
    assert isinstance(save, bool)
    assert isinstance(dims, bool)
    assert isinstance(dtypes, bool)
    assert isinstance(sum_stats, bool)

    graph = Digraph(strict=True)
    graph.node_attr.update(fontsize="12")

    _add_node(fin, graph, names=names, dims=dims, dtypes=dtypes, sum_stats=sum_stats)

    if save:
        graph.render(filename="computational_graph", cleanup=True)

    if render:
        return graph


def _add_node(node, graph, op_id=None, **kwargs):
    """ Recursively traces computational graph and adds nodes to Digraph. """
    node_id = str(id(node))
    node_lab = repr(node)
    if kwargs["names"] is not None:
        for key in kwargs["names"]:
            if id(kwargs["names"][key]) == id(node):
                node_lab = key
                break
            elif id(kwargs["names"][key]) == id(node.data):
                node_lab = key + "\n*Constant*"
                node_id = str(id(node.data))
                break
        if node_lab == repr(node):
            if not node.ndim:
                node_lab = str(node.data)
            elif node._constant:
                node_lab = "*Constant*"
            else:
                node_lab = "Intermediary Tensor"

    if node.ndim:
        if kwargs["dims"]:
            node_lab += "\nDims: {}".format(node.shape)
        if kwargs["dtypes"]:
            node_lab += "\nDtype: {}".format(node.dtype)
        if kwargs["sum_stats"]:
            node_lab += "\nMin: {min}\nMedian: {med}\nMean: {mean}\nMax: {max}".format(
                min=np.amin(node.data),
                med=np.median(node.data),
                mean=np.mean(node.data),
                max=np.amax(node.data),
            )
    else:
        if kwargs["dtypes"]:
            node_lab += "\nDtype: {}".format(node.dtype)

    graph.node(name=node_id, label=node_lab)

    if node._creator is None:
        if op_id is not None:
            graph.edge(node_id, op_id)
        return
    else:
        op_lab = repr(node._creator).rpartition(".")[-1].split(" ")[0]
        if op_id is not None:
            graph.edge(node_id, op_id)
        op_id = str(id(node._creator))

        graph.node(name=op_id, label=op_lab, style="filled", fillcolor="red")
        graph.edge(op_id, node_id)

        for var in node._creator.variables:
            _add_node(var, graph, op_id=op_id, **kwargs)



from numbers import Real

import numpy as np

from mygrad.math._special import logsumexp
from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor

__all__ = ["multiclass_hinge", "softmax_crossentropy", "margin_ranking_loss"]


def _check_loss_inputs(x, y_true):
    """ Ensures that the inputs to scores-truth style loss functions
    are of the correct shapes and types.

    Parameters
    ----------
    x : mygrad.Tensor, shape=(N, C)
        The C class scores for each of the N pieces of data.

    y_true : Sequence[int]
        The correct class-indices, in [0, C), for each datum.

    Raises
    ------
    TypeError
        `y_true` must be an integer-type array-like object
    ValueError
        `x` must be a 2-dimensional array-like object
        `y_true` must be a shape-(N,) array-like object
    """
    if not x.ndim == 2:
        raise ValueError(
            "`x` must be a 2-dimensional array-like object, got {}-dim".format(x.ndim)
        )

    if isinstance(y_true, Tensor):
        y_true = y_true.data

    y_true = np.asarray(y_true)
    if not np.issubdtype(y_true.dtype, np.integer):
        raise TypeError(
            "`y_true` must be an integer-type "
            "array-like object, got {}".format(y_true.dtype)
        )

    if y_true.ndim != 1 or y_true.shape[0] != x.shape[0]:
        raise ValueError(
            "`y_true` must be a shape-(N,) array: \n"
            "\tExpected shape-{}\n"
            "\tGot shape-{}".format((x.shape[0],), y_true.shape)
        )


class MulticlassHinge(Operation):
    def __call__(self, a, y, hinge=1.0):
        """ Computes the average multiclass hinge loss

        Parameters
        ----------
        a : mygrad.Tensor, shape=(N, C)
            The C class scores for each of the N pieces of data.

        y : numpy.ndarray, shape=(N,)
            The correct class-index, in [0, C), for each datum.

        Returns
        -------
        The average multiclass hinge loss

        Raises
        ------
        TypeError
            `y_true` must be an integer-type array-like object

        ValueError
            `x` must be a 2-dimensional array-like object
            `y_true` must be a shape-(N,) array-like object"""

        _check_loss_inputs(a, y)
        self.variables = (a,)
        scores = a.data
        correct_labels = (range(len(y)), y)
        correct_class_scores = scores[correct_labels]  # Nx1

        M = scores - correct_class_scores[:, np.newaxis] + hinge  # NxC margins
        not_thresh = np.where(M <= 0)
        Lij = M
        Lij[not_thresh] = 0
        Lij[correct_labels] = 0

        TMP = np.ones(M.shape, dtype=float)
        TMP[not_thresh] = 0
        TMP[correct_labels] = 0  # NxC; 1 where margin > 0
        TMP[correct_labels] = -1 * TMP.sum(axis=-1)
        self.back = TMP
        self.back /= scores.shape[0]
        return np.sum(Lij) / scores.shape[0]

    def backward_var(self, grad, index, **kwargs):
        return grad * self.back


def multiclass_hinge(x, y_true, hinge=1.0, constant=False):
    """ Parameters
        ----------
        x : array_like, shape=(N, K)
            The K class scores for each of the N pieces of data.

        y_true : array_like, shape=(N,)
            The correct class-indices, in [0, K), for each datum.

        hinge : float
            The size of the "hinge" outside of which a nonzero loss
            is incurred.

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        The average multiclass hinge loss

        Raises
        ------
        TypeError
            `y_true` must be an integer-type array-like object

        ValueError
            `x` must be a 2-dimensional array-like object
            `y_true` must be a shape-(N,) array-like object
        """
    return Tensor._op(MulticlassHinge, x, op_args=(y_true, hinge), constant=constant)


class SoftmaxCrossEntropy(Operation):
    """ Given the classification scores of C classes for N pieces of data,
        computes the NxC softmax classification probabilities. The
        cross entropy is then computed by using the true classification labels.

        log-softmax is used for improved numerical stability"""

    def __call__(self, x, y_true):
        """ Parameters
            ----------
            x : mygrad.Tensor, shape=(N, C)
                The C class scores for each of the N pieces of data.

            y_true : Sequence[int]
                The correct class-indices, in [0, C), for each datum.

            Returns
            -------
            The average softmax loss"""

        _check_loss_inputs(x, y_true)
        self.variables = (x,)
        scores = x.data
        log_softmax = scores - logsumexp(scores, axis=-1, keepdims=True)
        label_locs = (range(len(scores)), y_true)
        loss = -np.sum(log_softmax[label_locs]) / scores.shape[0]

        self.back = np.exp(log_softmax)
        self.back[label_locs] -= 1.0
        self.back /= scores.shape[0]
        return loss

    def backward_var(self, grad, index, **kwargs):
        return grad * self.back


def softmax_crossentropy(x, y_true, constant=False):
    r""" Given the classification scores of C classes for N pieces of data,

    computes the NxC softmax classification probabilities. The
    cross entropy is then computed by using the true classification labels.

    log-softmax is used for improved numerical stability.

    Parameters
    ----------
    x : array_like, shape=(N, C)
        The C class scores for each of the N pieces of data.

    y_true : array_like, shape=(N,)
        The correct class-indices, in [0, C), for each datum.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    The average softmax loss

    Raises
    ------
    ValueError
        Bad dimensionalities for ``x`` or ``y_true``

    Notes
    -----
    - :math:`N` is the number of samples in the batch.
    - :math:`C` is the number of possible classes for which scores are provided.

    Given the shape-:math:`(N, C)` tensor of scores, ``x``, the softmax classification
    probabilities are computed. That is, the score for class-:math:`k` of a given datum
    (:math:`s_{k}`) is normalized using the 'softmax' transformation:

    .. math::

        p_{k} = \frac{e^{s_k}}{\sum_{i=1}^{C}{e^{s_i}}}

    This produces the "prediction probability distribution", :math:`p`, for each datum.
    The cross-entropy loss for that datum is then computed according to the true class-index
    for that datum, as reported in ``y_true``. That is the "true probability distribution",
    :math:`t`, for the datum is :math:`1` for the correct class-index and :math:`0` elsewhere.

    The cross-entropy loss for that datum is thus:

    .. math::
       l = - \sum_{k=1}^{C}{t_{k} \log{p_{k}}}

    Having computed each per-datum cross entropy loss, this function then returns the loss
    averaged over all :math:`N` pieces of data:

    .. math::

       L = \frac{1}{N}\sum_{i=1}^{N}{l_{i}}

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet import softmax_crossentropy

    Let's take a simple case where N=1, and C=3. We'll thus make up classification
    scores for a single datum. Suppose the scores are identical for the three classes
    and that the true class is class-0:

    >>> x = mg.Tensor([[2., 2., 2.]])  # a shape-(1, 3) tensor of scores
    >>> y_true = mg.Tensor([0])  # the correct class for this datum is class-0

    Because the scores are identical for all three classes, the softmax normalization
    will simply produce :math:`p = [\frac{1}{3}, \frac{1}{3}, \frac{1}{3}]`. Because
    class-0 is the "true" class, :math:`t = [1., 0., 0.]`. Thus our softmax cross-entropy
    loss should be:

    .. math::
      -(1 \times \log{\frac{1}{3}} + 0 \times \log{\frac{1}{3}} + 0 \times \log{\frac{1}{3}})
      = \log(3) \approx 1.099

    Let's see that this is what ``softmax_crossentropy`` returns:

    >>> softmax_crossentropy(x, y_true)
    Tensor(1.09861229)

    Similarly, suppose a datum's scores are :math:`[0, 0, 10^6]`, then the softmax normalization
    will return :math:`p \approx [0., 0., 1.]`. If the true class for this datum is class-2, then
    the loss should be nearly 0, since :math:`p` and :math:`t` are essentially identical:

    .. math::
      -(0 \times \log{0} + 0 \times \log{0} + 1 \times \log{1})
      = -\log(1) = 0

    Now, let's construct ``x`` and ``y_true`` so that they incorporate the scores/labels for
    both of the data that we have considered:

    >>> x = mg.Tensor([[2., 2.,  2.],  # a shape-(2, 3) tensor of scores
    ...                [0., 0., 1E6]])
    >>> y_true = mg.Tensor([0, 2])     # the class IDs for the two data

    ``softmax_crossentropy(x, y_true)`` will return the average loss of these two data,
    :math:`\frac{1}{2}(1.099 + 0) \approx 0.55`:

    >>> softmax_crossentropy(x, y_true)
    Tensor(0.54930614)
    """
    return Tensor._op(SoftmaxCrossEntropy, x, op_args=(y_true,), constant=constant)


class MarginRanking(Operation):
    def __call__(self, x1, x2, y, margin):
        """ Computes the margin ranking loss between ``x1``
        and ``x2``.

        Parameters
        ----------
        x1 : mygrad.Tensor, shape=(N,) or (N, D)

        x2 : mygrad.Tensor, shape=(N,) or (N, D)

        y : numpy.ndarray

        margin : float

        Returns
        -------
        numpy.ndarray, shape=()
        """
        self.variables = (x1, x2)
        x1 = x1.data
        x2 = x2.data

        self.y = y

        M = margin - self.y * (x1 - x2)
        not_thresh = M <= 0
        loss = M
        loss[not_thresh] = 0.0

        self._grad = np.ones_like(M)
        self._grad[not_thresh] = 0.0
        self._grad /= M.size
        return np.mean(loss)

    def backward_var(self, grad, index, **kwargs):
        sign = -self.y if index == 0 else self.y
        return grad * (sign * self._grad)


def margin_ranking_loss(x1, x2, y, margin, constant=False):
    r"""Computes the margin average margin ranking loss.
    Equivalent to::

    >>> import mygrad as mg
    >>> mg.mean(mg.maximum(0, margin - y * (x1 - x2)))

    Parameters
    ----------
    x1 : array_like, shape=(N,) or (N, D)
        A batch of scores or descriptors to compare against those in `x2`

    x2 : array_like, shape=(N,) or (N, D)
        A batch of scores or descriptors to compare against those in `x1`

    y  : Union[int, array_like], scalar or shape=(N,)
        1 or -1. Specifies whether the margin is compared against `(x1 - x2)`
        or `(x2 - x1)`, for each of the N comparisons.

    margin : float
        A non-negative value to be used as the margin for the loss.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor, shape=()
        The mean margin ranking loss.
    """
    if not 0 < x1.ndim < 3:
        raise ValueError("`x1` must have shape (N,) or (N, D)")
    if not x1.shape == x2.shape:
        raise ValueError("`x1` and `x2` must have the same shape")
    if not np.issubdtype(x1.dtype, np.floating):
        raise TypeError("`x1` must contain floats")
    if not np.issubdtype(x2.dtype, np.floating):
        raise TypeError("`x2` must contain floats")
    if not isinstance(margin, Real) or margin < 0:
        raise ValueError("`margin` must be a non-negative scalar")

    if isinstance(y, Tensor):
        y = y.data

    y = np.asarray(y)

    if y.size == 1:
        y = np.array(y.item())

    if not y.ndim == 0 and not (y.ndim == 1 and len(y) == len(x1)):
        raise ValueError("`y` must be a scalar or shape-(N,) array of ones")

    if y.ndim:
        if x1.ndim == 2:
            y = y.reshape(-1, 1)
    return Tensor._op(MarginRanking, x1, x2, op_args=(y, margin), constant=constant)



from mygrad.tensor_base import Tensor

from .ops import Abs, Cbrt, Maximum, Minimum, Sqrt

__all__ = ["abs", "absolute", "cbrt", "sqrt", "maximum", "minimum"]


def abs(a, constant=False):
    """ ``f(a) -> abs(a)``

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not back-propagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Notes
        -----
        The derivative at a == 0 returns nan"""
    return Tensor._op(Abs, a, constant=constant)


absolute = abs


def sqrt(a, constant=False):
    """ ``f(a) -> sqrt(a)``

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not backpropagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Sqrt, a, constant=constant)


def cbrt(a, constant=False):
    """ ``f(a) -> cbrt(a)``

        Parameters
        ----------
        a : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not backpropagate a gradient)

        Returns
        -------
        mygrad.Tensor"""
    return Tensor._op(Cbrt, a, constant=constant)


def maximum(a, b, constant=False):
    """ Element-wise maximum of array elements.

        Parameters
        ----------
        a : array_like

        b : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not backpropagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Notes
        -----
        The gradient does not exist where a == b; we use a
        value of 0 here."""
    return Tensor._op(Maximum, a, b, constant=constant)


def minimum(a, b, constant=False):
    """ Element-wise minimum of array elements.

        Parameters
        ----------
        a : array_like

        b : array_like

        constant : bool, optional(default=False)
            If ``True``, the returned tensor is a constant (it
            does not backpropagate a gradient)

        Returns
        -------
        mygrad.Tensor

        Notes
        -----
        The gradient does not exist where a == b; we use a
        value of 0 here."""
    return Tensor._op(Minimum, a, b, constant=constant)



from numbers import Integral

import numpy as np

from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor

try:
    from numba import njit, vectorize
except ImportError:  # pragma: no cover
    raise ImportError(
        "The package `numba` must be installed in order to access the gru."
    )


@vectorize(
    ["int32(int32)", "int64(int64)", "float32(float32)", "float64(float64)"],
    nopython=True,
)
def sig(f):  # pragma: no cover
    """
    Calculates a sigmoid function
    """
    return 1 / (1 + np.exp(-f))


@vectorize(
    ["int32(int32)", "int64(int64)", "float32(float32)", "float64(float64)"],
    nopython=True,
)
def d_sig(f):  # pragma: no cover
    """
    Calculates the derivative of a sigmoid function
    """
    return f * (1 - f)


@vectorize(
    ["int32(int32)", "int64(int64)", "float32(float32)", "float64(float64)"],
    nopython=True,
)
def d_tanh(f):  # pragma: no cover
    """
    Calculates the derivative of a tanh function
    """
    return 1 - f ** 2


@njit
def dot(a, b):
    """
    Calculates the dot product between 2 arrays
    of shapes (W,X,Y) and (Y,Z), respectively
    """
    out = np.zeros((a.shape[0], a.shape[1], b.shape[-1]))
    for i in range(len(a)):
        out[i] = np.dot(a[i], b)
    return out


@njit
def _gru_layer(s, z, r, h, Wz, Wr, Wh):
    """ Given:
            S(t=0)
            z = X(t) Uz + bz
            r = X(t) Ur + br
            h = X(t) Uh + bh

        Compute Z(t), R(t), H(t), S(t) for all 1 <= t <= T

        Parameters
        ----------
        s : numpy.ndarray, shape=(T+1, N, D)
            Modified in-place
        z : numpy.ndarray, shape=(T, N, D)
            Modified in-place
        r : numpy.ndarray, shape=(T, N, D)
            Modified in-place
        h : numpy.ndarray, shape=(T, N, D)
            Modified in-place
        Wz : numpy.ndarray, shape=(D, D)
        Wr : numpy.ndarray, shape=(D, D)
        Wh : numpy.ndarray, shape=(D, D) """
    for n in range(len(s) - 1):
        z[n] += np.dot(s[n], Wz)
        z[n] = sig(z[n])

        r[n] += np.dot(s[n], Wr)
        r[n] = sig(r[n])

        h[n] += np.dot(r[n] * s[n], Wh)
        h[n] = np.tanh(h[n])

        s[n + 1] = (1 - z[n]) * h[n] + z[n] * s[n]


@njit
def _gru_dLds(s, z, r, dLds, Wz, Wh, Wr, dz, dh, dr, s_h, one_z):
    """
                        Z_{t} = sigmoid(Uz X_{t} + Wz S_{t-1} + bz)
                        R_{t} = sigmoid(Ur X_{t} + Wr S_{t-1} + br)
                        H_{t} = tanh(Uh X_{t} + Wh (R{t} * S_{t-1}) + bh)
                        S_{t} = (1 - Z{t}) * H{t} + Z{t} * S_{t-1}

    Returns
    --------

        dL / ds(t) =   partial dL / ds(t+1) * ds(t+1) / ds(t)
                     + partial dL / ds(t+1) * ds(t+1) / dz(t) * dz(t) / ds(t)
                     + partial dL / ds(t+1) * ds(t+1) / dh(t) * dh(t) / ds(t)
                     + partial dL / ds(t+1) * ds(t+1) / dh(t) * dh(t) / dr(t) * dr(t) / ds(t)
    """
    dLdh = dot(dLds * one_z * dh, Wh)

    out = z * dLds
    out += dot(dLds * s_h * dz, Wz)
    out += dLdh * r
    out += dot(dLdh * s * dr, Wr)

    return out


@njit
def _gru_bptt(
    X, dLds, s, z, r, Wz, Wh, Wr, dz, dh, dr, s_h, one_z, bp_lim, old_dLds=None
):
    Wz, Wh, Wr = Wz.T, Wh.T, Wr.T
    bptt = bp_lim < len(X) - 1
    if bptt:  # pragma: no cover
        old_dLds = np.zeros_like(dLds)

    for i in range(bp_lim):
        #  dL(t) / ds(t) + dL(t+1) / ds(t)
        if bptt:  # pragma: no cover
            source_index = slice(1, len(dLds) - i)
            target_index = slice(None, len(dLds) - (i + 1))
            dt = dLds[source_index] - old_dLds[source_index]
            old_dLds = np.copy(dLds)
        else:  # no backprop truncation
            source_index = slice(len(dLds) - (i + 1), len(dLds) - i)
            target_index = slice(len(dLds) - (i + 2), len(dLds) - (i + 1))
            dt = dLds[source_index]

        dLds[target_index] += _gru_dLds(
            s[source_index],
            z[source_index],
            r[source_index],
            dt,
            Wz,
            Wh,
            Wr,
            dz[source_index],
            dh[source_index],
            dr[source_index],
            s_h[source_index],
            one_z[source_index],
        )


def _backprop(var, grad):  # pragma: no cover
    if not var.constant:
        if var.grad is None:
            var.grad = np.asarray(grad)
        else:
            var.grad += grad


class GRUnit(Operation):
    scalar_only = True

    def __call__(
        self, X, Uz, Wz, bz, Ur, Wr, br, Uh, Wh, bh, s0=None, bp_lim=None, dropout=0.0
    ):
        if bp_lim is not None:
            assert isinstance(bp_lim, Integral) and 0 <= bp_lim < len(X)
        assert 0.0 <= dropout < 1.0
        self._dropout = dropout
        self.bp_lim = bp_lim if bp_lim is not None else len(X) - 1

        self.X = X  # type: Tensor  # shape=(T, N, C)

        self.Uz = Uz  # type: Tensor  # shape=(C, D)
        self.Wz = Wz  # type: Tensor  # shape=(D, D)
        self.bz = bz  # type: Tensor  # shape=(D,)

        self.Ur = Ur  # type: Tensor  # shape=(C, D)
        self.Wr = Wr  # type: Tensor  # shape=(D, D)
        self.br = br  # type: Tensor  # shape=(D,)

        self.Uh = Uh  # type: Tensor  # shape=(C, D)
        self.Wh = Wh  # type: Tensor  # shape=(D, D)
        self.bh = bh  # type: Tensor  # shape=(D,)

        self.variables = (
            self.X,
            self.Uz,
            self.Wz,
            self.bz,
            self.Ur,
            self.Wr,
            self.br,
            self.Uh,
            self.Wh,
            self.bh,
        )
        T, N, C = X.shape
        D, = bz.shape

        seq = self.X.data

        # t starts at 0 for S; all other sequences begin at t = 1
        out = np.zeros((T + 1, N, D))

        if s0 is not None:
            out[0] = s0.data if isinstance(s0, Tensor) else s0

        # compute all contributions to Z, R, H from the input sequence
        # shape: T, N, D
        z = np.tensordot(seq, self.Uz.data, [[-1], [0]])
        r = np.tensordot(seq, self.Ur.data, [[-1], [0]])
        h = np.tensordot(seq, self.Uh.data, [[-1], [0]])

        if dropout:
            p = 1 - dropout
            # For Uz/Ur/Uh: a dropout mask is generated for each datum and is applied uniformly across T
            self._dropUz, self._dropUr, self._dropUh = (
                np.random.binomial(1, p, size=(3, 1, N, D)) / p
            )
            self._dropWz, self._dropWr, self._dropWh = (
                np.random.binomial(1, p, size=(3, D, D)) / p
            )

            z *= self._dropUz
            r *= self._dropUr
            h *= self._dropUh

            Wz = self._dropWz * self.Wz.data
            Wr = self._dropWr * self.Wr.data
            Wh = self._dropWh * self.Wh.data

        else:
            self._dropUz, self._dropUr, self._dropUh = None, None, None
            self._dropWz, self._dropWr, self._dropWh = None, None, None
            Wz = self.Wz.data
            Wr = self.Wr.data
            Wh = self.Wh.data

        z += bz.data  # X Uz + bz
        r += br.data  # X Ur + br
        h += bh.data  # X Uh + bh

        _gru_layer(out, z, r, h, Wz, Wr, Wh)

        self._hidden_seq = Tensor(out, _creator=self)
        self._z = Tensor(z, _creator=self)
        self._r = Tensor(r, _creator=self)
        self._h = Tensor(h, _creator=self)

        return self._hidden_seq

    def backward(self, grad, *, graph, **kwargs):

        s = self._hidden_seq.data[:-1]
        z = self._z.data
        r = self._r.data
        h = self._h.data

        dLds = grad[1:]

        const = {"1 - h**2": d_tanh(h), "z*(1 - z)": d_sig(z), "r*(1 - r)": d_sig(r)}

        if self._dropout:
            Wz = self._dropWz * self.Wz.data
            Wr = self._dropWr * self.Wr.data
            Wh = self._dropWh * self.Wh.data
        else:
            Wz = self.Wz.data
            Wr = self.Wr.data
            Wh = self.Wh.data

        const["s - h"] = s - h
        const["1 - z"] = 1 - z

        _gru_bptt(
            self.X.data,
            dLds,
            s,
            z,
            r,
            Wz,
            Wh,
            Wr,
            const["z*(1 - z)"],
            const["1 - h**2"],
            const["r*(1 - r)"],
            const["s - h"],
            const["1 - z"],
            self.bp_lim,
        )

        zgrad = dLds * const["s - h"]  # dL / dz
        hgrad = dLds * const["1 - z"]  # dL / dh
        rgrad = dot(const["1 - h**2"] * hgrad, Wh.T) * s  # dL / dr

        self._hidden_seq.grad = dLds

        if not (self.Uz.constant and self.Wz.constant and self.bz.constant):
            dz = zgrad * const["z*(1 - z)"]
        # backprop through Wz
        if not self.Wz.constant:
            dWz = np.tensordot(s, dz, ([0, 1], [0, 1]))
            if self._dropout:
                dWz *= self._dropWz
            _backprop(self.Wz, dWz)  # self.Wz.backward(dWz, **kwargs)
        # backprop through bz
        if not self.bz.constant:
            _backprop(self.bz, dz.sum(axis=(0, 1)))
        # backprop through bz
        if not self.Uz.constant:
            if self._dropout:
                dz *= (
                    self._dropUz
                )  # IMPORTANT augmented update: this must come after Wz and bz backprop
            _backprop(self.Uz, np.tensordot(self.X.data, dz, ([0, 1], [0, 1])))

        if not (self.Ur.constant and self.Wr.constant and self.br.constant):
            dr = rgrad * const["r*(1 - r)"]
        # backprop through Wr
        if not self.Wr.constant:
            dWr = np.tensordot(s, dr, ([0, 1], [0, 1]))
            if self._dropout:
                dWr *= self._dropWr
            _backprop(self.Wr, dWr)
        # backprop through br
        if not self.br.constant:
            _backprop(
                self.br, dr.sum(axis=(0, 1))
            )  # self.br.backward(dr.sum(axis=(0, 1)), **kwargs)
        # backprop through Ur
        if not self.Ur.constant:
            if self._dropout:
                dr *= (
                    self._dropUr
                )  # IMPORTANT augmented update: this must come after Wr and br backprop
            _backprop(self.Ur, np.tensordot(self.X.data, dr, ([0, 1], [0, 1])))

        if not (self.Uh.constant and self.Wh.constant and self.bh.constant):
            dh = hgrad * const["1 - h**2"]
        # backprop through Wh
        if not self.Wh.constant:
            dWh = np.tensordot((s * r), dh, ([0, 1], [0, 1]))
            if self._dropout:
                dWh *= self._dropWh
            _backprop(self.Wh, dWh)  # self.Wh.backward(dWh, **kwargs)
        # backprop through bh
        if not self.bh.constant:
            _backprop(
                self.bh, dh.sum(axis=(0, 1))
            )  # self.bh.backward(dh.sum(axis=(0, 1)), **kwargs)
        # backprop through Uh
        if not self.Uh.constant:
            if self._dropout:
                dh *= (
                    self._dropUh
                )  # IMPORTANT augmented update: this must come after Wh and bh backprop
            _backprop(self.Uh, np.tensordot(self.X.data, dh, ([0, 1], [0, 1])))

        # backprop through X
        if not self.X.constant:
            tmp = dLds * const["1 - z"] * const["1 - h**2"]
            if not self._dropout:
                dLdX = dot((dLds * const["s - h"]) * const["z*(1 - z)"], self.Uz.data.T)
                dLdX += dot(tmp, self.Uh.data.T)
                dLdX += dot(dot(tmp, Wh.T) * s * const["r*(1 - r)"], self.Ur.data.T)
            else:
                dLdX = dot(
                    (self._dropUz * (dLds * const["s - h"]) * const["z*(1 - z)"]),
                    self.Uz.data.T,
                )
                dLdX += dot(self._dropUh * tmp, self.Uh.data.T)
                dLdX += dot(
                    self._dropUr * (dot(tmp, Wh.T) * s * const["r*(1 - r)"]),
                    self.Ur.data.T,
                )
            _backprop(self.X, dLdX)  # self.X.backward(dLdX, **kwargs)

        del self._z
        del self._r
        del self._h

        for x in (
            self.X,
            self.Uz,
            self.Wz,
            self.bz,
            self.Ur,
            self.Wr,
            self.br,
            self.Uh,
            self.Wh,
            self.bh,
        ):
            if not x.constant:
                x._accum_ops.add(self)
                x._backward(graph=graph)


def gru(
    X,
    Uz,
    Wz,
    bz,
    Ur,
    Wr,
    br,
    Uh,
    Wh,
    bh,
    s0=None,
    bp_lim=None,
    dropout=0.0,
    constant=False,
):
    r""" Performs a forward pass of sequential data through a Gated Recurrent Unit layer, returning
    the 'hidden-descriptors' arrived at by utilizing the trainable parameters as follows::

                Z_{t} = sigmoid(X_{t} Uz + S_{t-1} Wz + bz)
                R_{t} = sigmoid(X_{t} Ur + S_{t-1} Wr + br)
                H_{t} =    tanh(X_{t} Uh + (R{t} * S_{t-1}) Wh + bh)
                S_{t} = (1 - Z{t}) * H{t} + Z{t} * S_{t-1}

    Parameters
    ----------
    X : array_like, shape=(T, N, C)
       The sequential data to be passed forward.

    Uz : array_like, shape=(C, D)
       The weights used to map sequential data to its hidden-descriptor representation

    Wz : array_like, shape=(D, D)
        The weights used to map a hidden-descriptor to a hidden-descriptor.

    bz : array_like, shape=(D,)
       The biases used to scale a hidden-descriptor.

    Ur : array_like, shape=(C, D)
       The weights used to map sequential data to its hidden-descriptor representation

    Wr : array_like, shape=(D, D)
        The weights used to map a hidden-descriptor to a hidden-descriptor.

    br : array_like, shape=(D,)
       The biases used to scale a hidden-descriptor.

    Uh : array_like, shape=(C, D)
       The weights used to map sequential data to its hidden-descriptor representation

    Wh : array_like, shape=(D, D)
        The weights used to map a hidden-descriptor to a hidden-descriptor.

    bh : array_like, shape=(D,)
       The biases used to scale a hidden-descriptor.

    s0 : Optional[array_like], shape=(N, D)
        The 'seed' hidden descriptors to feed into the RNN. If None, a Tensor
        of zeros of shape (N, D) is created.

    bp_lim : Optional[int]
        *This feature is experimental and is currently untested*.
        The (non-zero) limit of the depth of back propagation through time to be
        performed. If `None` back propagation is passed back through the entire sequence.

        E.g. `bp_lim=3` will propagate gradients only up to 3 steps backward through the
        recursive sequence.

    dropout : float (default=0.), 0 <= dropout < 1
        If non-zero, the dropout scheme described in [1]_ is applied. See Notes
        for more details.

    constant : bool, optional (default=False)
        If True, the resulting Tensor is a constant.

    Returns
    -------
    mygrad.Tensor, shape=(T+1, N, D)
        The sequence of 'hidden-descriptors' produced by the forward pass of the RNN.

    Notes
    -----
    - :math:`T` : Sequence length
    - :math:`N` : Batch size
    - :math:`C` : Length of single datum
    - :math:`D` : Length of 'hidden' descriptor

    The GRU system of equations is given by:

    .. math::

                Z_{t} = \sigma (X_{t} U_z + S_{t-1} Wz + bz)

                R_{t} = \sigma (X_{t} U_r + S_{t-1} Wr + br)

                H_{t} =    tanh(X_{t} U_h + (R_{t} * S_{t-1}) W_h + b_h)

                S_{t} = (1 - Z_{t}) * H_{t} + Z_{t} * S_{t-1}

    Following the dropout scheme specified in [1]_, the hidden-hidden weights (Wz/Wr/Wh)
    randomly have their weights dropped prior to forward/back-prop. The input connections
    (via Uz/Ur/Uh) have variational dropout ([2]_) applied to them with a common dropout
    mask across all t. That is three static dropout masks, each with shape-(N,D), are
    applied to

    .. math::
                                          X_{t} U_z

                                          X_{t} U_r

                                          X_{t} U_h
    respectively, for all :math:`t`.

    References
    ----------
    .. [1] S. Merity, et. al. "Regularizing and Optimizing LSTM Language Models",
           arXiv:1708.02182v1, 2017.

    .. [2] Y. Gal, Z. Ghahramani "A Theoretically Grounded Application of Dropout
           in Recurrent Neural Networks" arXiv:1512.05287v5, 2016. """
    if s0 is not None:
        if not isinstance(s0, np.ndarray) and not (
            isinstance(s0, Tensor) and (constant or s0.constant)
        ):
            raise ValueError(
                "GRU does not support non-constant tensors for the initial hidden"
                "state value, `s0`"
            )
    s = Tensor._op(
        GRUnit,
        X,
        Uz,
        Wz,
        bz,
        Ur,
        Wr,
        br,
        Uh,
        Wh,
        bh,
        op_kwargs=dict(s0=s0, bp_lim=bp_lim, dropout=dropout),
        constant=constant,
    )
    s.creator._hidden_seq = s
    return s



import numpy as np

__all__ = [
    'Inf',
    'Infinity',
    'NAN',
    'NINF',
    'NZERO',
    'NaN',
    'PINF',
    'PZERO',
    'e',
    'euler_gamma',
    'inf',
    'infty',
    'nan',
    'newaxis',
    'pi',
]

Inf = np.Inf
Infinity = np.Infinity
NAN = np.NAN
NINF = np.NINF
NZERO = np.NZERO
NaN = np.NaN
PINF = np.PINF
PZERO = np.PZERO
e = np.e
euler_gamma = np.euler_gamma
inf = np.inf
infty = np.infty
nan = np.nan
newaxis = np.newaxis
pi = np.pi



from setuptools import find_packages, setup

import versioneer

DISTNAME = "mygrad"
LICENSE = "MIT"
AUTHOR = "Ryan Soklaski"
AUTHOR_EMAIL = "rsoklaski@gmail.com"
URL = "https://github.com/rsokl/MyGrad"
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Intended Audience :: Education",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering",
]

INSTALL_REQUIRES = ["numpy >= 1.12"]
TESTS_REQUIRE = ["pytest >= 3.8", "hypothesis >= 4.6", "scipy"]

DESCRIPTION = "A sleek auto-differentiation library that wraps numpy."
LONG_DESCRIPTION = """
mygrad is a simple, NumPy-centric autograd library. An autograd library enables
you to automatically compute derivatives of mathematical functions. This library is
designed to serve primarily as an education tool for learning about gradient-based
machine learning; it is easy to install, has a readable and easily customizable code base,
and provides a sleek interface that mimics NumPy. Furthermore, it leverages NumPy's
vectorization to achieve good performance despite the library's simplicity.

This is not meant to be a competitor to libraries like PyTorch (which mygrad most
closely resembles) or TensorFlow. Rather, it is meant to serve as a useful tool for
students who are learning about training neural networks using back propagation.
"""


setup(
    name=DISTNAME,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license=LICENSE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    classifiers=CLASSIFIERS,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    extras_require={
        "rnn": ["numba>=0.34.0"]  # GRU and vanilla RNN require numba-acceleration
    },
    url=URL,
    download_url="https://github.com/rsokl/mygrad/tarball/" + versioneer.get_version(),
    python_requires=">=3.5",
    packages=find_packages(exclude=["tests", "tests.*"]),
)



from numbers import Integral

import numpy as np
from numpy.lib.stride_tricks import as_strided


def sliding_window_view(arr, window_shape, step, dilation=None):
    """ Create a sliding window view over the trailing dimensions of an array.
        No copy is made.

        The window is applied only to valid regions of ``arr``, but is applied geedily.

        See Notes section for details.

        Parameters
        ----------
        arr : numpy.ndarray, shape=(..., [x, (...), z])
            C-contiguous array over which sliding view-window is applied along the trailing
            dimensions ``[x, ..., z]``, as determined by the length of ``window_shape``.

            If ``arr`` is not C-contiguous, it will be replaced by ``numpy.ascontiguousarray(arr)``

        window_shape : Sequence[int]
            Specifies the shape of the view-window: ``[Wx, (...), Wz]``.
            The length of `window_shape` determines the length of ``[x, (...) , z]``.

        step : Union[int, Sequence[int]]
            The step sized used along the ``[x, (...), z]`` dimensions: ``[Sx, (...), Sz]``.
            If a single integer is specified, a uniform step size is used.

        dilation : Optional[Union[int, Sequence[int]]]
            The dilation factor used along the ``[x, (...), z]`` directions: ``[Dx, (...), Dz]``.
            If no value is specified, a dilation factor of 1 is used along each direction.
            Dilation specifies the step size used when filling the window's elements

        Returns
        -------
        numpy.ndarray
            A contiguous view of ``arr``, of shape ``([X, (...), Z], ..., [Wx, (...), Wz])``, where
            ``[X, ..., Z]`` is the shape of the grid on which the window was applied. See Notes
            sections for more details.

        Raises
        ------
        ValueError, TypeError
            Invalid step-size, window shape, or dilation

        Notes
        -----
        Window placement:
            Given a dimension of size x, with a window of size W along this dimension, applied
            with stride S and dilation D, the window will be applied::
                                      X = (x - (W - 1) * D + 1) // S + 1
            number of times along that dimension.

        Interpreting output:
            In general, given an array ``arr`` of shape (..., x, (...), z), and::

                out = sliding_window_view(arr, window_shape=[Wx, (...), Wz], step=[Sx, (...), Sz])

            then indexing ``out`` with ``[ix, (...), iz]`` produces the following view of ``x``::

                out[ix, (...), iz] ==
                    x[..., ix*Sx:(ix*Sx + Wx*Dx):Dx, (...), iz*Sz:(iz*Sz + Wz*Dz):Dz]

            For example, suppose ``arr`` is an array of shape-(10, 12, 6). Specifying sliding
            window of shape ``(3, 3)`` with step size ``(2, 2)``, dilation ``(2, 1)`` will create the view::

                            [[arr[:,  0:6:2, 0:3], arr[:,   0:6:3, 3:6]]
                             [arr[:, 6:12:2, 0:3], arr[:, 6:12:12, 3:6]]]

            producing a view of shape ``(2, 2, 10, 3, 3)`` in total.

        Examples
        --------
        >>> import numpy as np
        >>> x = np.arange(36).reshape(6, 6)
        >>> x
        array([[ 0,  1,  2,  3,  4,  5],
               [ 6,  7,  8,  9, 10, 11],
               [12, 13, 14, 15, 16, 17],
               [18, 19, 20, 21, 22, 23],
               [24, 25, 26, 27, 28, 29],
               [30, 31, 32, 33, 34, 35]])

        Apply an 3x2 window with step-sizes of (2, 2). This results in
        the window being placed twice along axis-0 and three times along axis-1.

        >>> y = sliding_window_view(x, step=(2, 2), window_shape=(3, 2))
        >>> y.shape
        (2, 3, 3, 2)

        window applied at (0, 0)

        >>> y[0, 0]
        array([[ 0,  1],
               [ 6,  7],
               [12, 13]])

        window applied at (2, 0)

        >>> y[1, 0]
        array([[12, 13],
               [18, 19],
               [24, 25]])

        window applied at (0, 2)

        >>> y[0, 1]
        array([[ 2,  3],
               [ 8,  9],
               [14, 15]])

        verify that an element in this window-view is correct

        >>> i, j = np.random.randint(0, 2, size=2)
        >>> wx, wy = (2, 2)
        >>> sx, sy = (2, 2)
        >>> np.all(y[i, j] == x[..., i*sx:(i*sx + wx), j*sy:(j*sy + wy)])
        True
        """

    if not hasattr(window_shape, "__iter__"):
        raise TypeError(
            "`window_shape` must be a sequence of positive integers, got: {}".format(
                window_shape
            )
        )
    window_shape = tuple(window_shape)
    if not all(isinstance(i, Integral) and i > 0 for i in window_shape):
        raise TypeError(
            "`window_shape` must be a sequence of positive integers, "
            "got: {}".format(window_shape)
        )

    if len(window_shape) > arr.ndim:
        raise ValueError(
            "`window_shape` ({}) cannot specify more values than "
            "`arr.ndim` ({}).".format(window_shape, arr.ndim)
        )

    if not isinstance(step, Integral) and not hasattr(step, "__iter__"):
        raise TypeError(
            "`step` must be a positive integer or a sequence of positive "
            "integers, got: {}".format(step)
        )

    step = (
        (int(step),) * len(window_shape) if isinstance(step, Integral) else tuple(step)
    )

    if not all(isinstance(i, Integral) and i > 0 for i in step):
        raise ValueError(
            "`step` must be a positive integer or a sequence of positive "
            "integers, got: {}".format(step)
        )

    if any(i > j for i, j in zip(window_shape[::-1], arr.shape[::-1])):
        raise ValueError(
            "Each size of the window-shape must fit within the trailing "
            "dimensions of `arr`."
            "{} does not fit in {}".format(
                window_shape, arr.shape[-len(window_shape) :]
            )
        )

    if (
        dilation is not None
        and not isinstance(dilation, Integral)
        and not hasattr(dilation, "__iter__")
    ):
        raise TypeError(
            "`dilation` must be None, a positive integer, or a sequence of "
            "positive integers, got: {}".format(dilation)
        )
    if dilation is None:
        dilation = np.ones((len(window_shape),), dtype=int)
    else:
        if isinstance(dilation, Integral):
            dilation = np.full((len(window_shape),), fill_value=dilation, dtype=int)
        else:
            np.asarray(dilation)

        if not all(isinstance(i, Integral) and i > 0 for i in dilation) or len(
            dilation
        ) != len(window_shape):
            raise ValueError(
                "`dilation` must be None, a positive integer, or a sequence of "
                "positive integers with the same length as `window_shape` "
                "({}), got: {}".format(window_shape, dilation)
            )
        if any(
            w * d > s
            for w, d, s in zip(window_shape[::-1], dilation[::-1], arr.shape[::-1])
        ):
            raise ValueError(
                "The dilated window ({}) must fit within the trailing "
                "dimensions of `arr` ({})".format(
                    tuple(w * d for w, d in zip(window_shape, dilation)),
                    arr.shape[-len(window_shape) :],
                )
            )

    if not arr.flags["C_CONTIGUOUS"]:
        arr = np.ascontiguousarray(arr)

    step = np.array(step)  # (Sx, ..., Sz)
    window_shape = np.array(window_shape)  # (Wx, ..., Wz)
    in_shape = np.array(arr.shape[-len(step) :])  # (x, ... , z)
    nbyte = arr.strides[-1]  # size, in bytes, of element in `arr`

    # per-byte strides required to fill a window
    win_stride = tuple(np.cumprod(arr.shape[:0:-1])[::-1]) + (1,)

    # per-byte strides required to advance the window
    step_stride = tuple(win_stride[-len(step) :] * step)

    # update win_stride to accommodate dilation
    win_stride = np.array(win_stride)
    win_stride[-len(step) :] *= dilation
    win_stride = tuple(win_stride)

    # tuple of bytes to step to traverse corresponding dimensions of view
    # see: 'internal memory layout of an ndarray'
    stride = tuple(int(nbyte * i) for i in step_stride + win_stride)

    # number of window placements along x-dim: X = (x - (Wx - 1)*Dx + 1) // Sx + 1
    out_shape = tuple((in_shape - ((window_shape - 1) * dilation + 1)) // step + 1)

    # ([X, (...), Z], ..., [Wx, (...), Wz])
    out_shape = out_shape + arr.shape[: -len(step)] + tuple(window_shape)
    out_shape = tuple(int(i) for i in out_shape)

    return as_strided(arr, shape=out_shape, strides=stride, writeable=False)



from functools import reduce

import numpy as np

from mygrad.operation_base import BroadcastableOp, Operation

__all__ = [
    "Add",
    "Subtract",
    "Multiply",
    "Divide",
    "Reciprocal",
    "Power",
    "Square",
    "Positive",
    "Negative",
    "AddSequence",
    "MultiplySequence",
]


class Add(BroadcastableOp):
    def __call__(self, a, b):
        """ Performs 'add' forward-pass: f(a,b) -> a + b

            Parameters
            ----------
            a : mygrad.Tensor
            b : mygrad.Tensor

            Returns
            -------
            out : numpy.ndarray """

        self.variables = (a, b)
        out = a.data + b.data
        return out

    def backward_var(self, grad, index, **kwargs):
        return grad


class Subtract(BroadcastableOp):
    def __call__(self, a, b):
        """ f(a,b) -> a - b

            Parameters
            ----------
            a : mygrad.Tensor
            b : mygrad.Tensor

            Returns
            -------
            out : numpy.ndarray """
        self.variables = (a, b)
        out = a.data - b.data
        return out

    def backward_var(self, grad, index, **kwargs):
        if index == 0:
            return grad
        else:
            return -grad


class Multiply(BroadcastableOp):
    def __call__(self, a, b):
        """ Parameters
            ----------
            a : mygrad.Tensor
            b : mygrad.Tensor"""
        self.variables = (a, b)
        out = a.data * b.data
        return out

    def backward_var(self, grad, index, **kwargs):
        a, b = self.variables
        if index == 0:  # backprop through a
            return grad * b.data
        elif index == 1:  # backprop through b
            return grad * a.data


class Divide(BroadcastableOp):
    def __call__(self, a, b):
        """ f(a, b) -> a / b"""
        self.variables = (a, b)
        out = a.data / b.data
        return out

    def backward_var(self, grad, index, **kwargs):
        a, b = self.variables
        if index == 0:  # backprop through a
            return grad / b.data
        else:  # broadcast through b
            return -grad * a.data / (b.data ** 2)


class Reciprocal(BroadcastableOp):
    def __call__(self, a):
        """ f(a) -> 1 / a"""
        self.variables = (a,)
        return np.reciprocal(a.data)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        return -grad * np.reciprocal(a.data ** 2)


class Power(BroadcastableOp):
    def __call__(self, a, b):
        """ f(a, b) -> a ** b

            Parameters
            ----------
            a: mygrad.Tensor
            b: mygrad.Tensor"""
        self.variables = (a, b)
        out = a.data ** b.data
        return out

    def backward_var(self, grad, index, **kwargs):
        a, b = self.variables
        x, y = a.data, b.data
        if index == 0:
            return grad * y * (x ** np.where(y, (y - 1), 1))
        else:
            return grad * (x ** y) * np.log(np.where(x, x, 1))


class Square(Operation):
    def __call__(self, a):
        """ f(a) -> a ** 2

            Parameters
            ----------
            a : mygrad.Tensor"""
        self.variables = (a,)
        return np.square(a.data)

    def backward_var(self, grad, index, **kwargs):
        return grad * 2 * self.variables[index].data


class Positive(Operation):
    """ f(a) = +a """

    def __call__(self, a, where=True):
        """
        Parameters
        ----------
        a: mygrad.Tensor

        where : array_like, optional
            Values of True indicate to calculate the ufunc at that position,
            values of False indicate to leave the value in the output alone."""
        self.variables = (a,)
        self.conf = dict(where=where)
        return np.positive(a.data, where=where)

    def backward_var(self, grad, index, **kwargs):
        return np.positive(grad, **self.conf)


class Negative(Operation):
    """ f(a) = -a """

    def __call__(self, a, where=True):
        """
        Parameters
        ----------
        a : mygrad.Tensor

        where : array_like, optional
            Values of True indicate to calculate the ufunc at that position,
            values of False indicate to leave the value in the output alone."""
        self.variables = (a,)
        self.conf = dict(where=where)
        return np.negative(a.data, where=where)

    def backward_var(self, grad, index, **kwargs):
        return np.negative(grad, **self.conf)


class AddSequence(BroadcastableOp):
    """Performs f(a, b, ..., z) = a + b + ... + z"""

    def __call__(self, *input_vars):
        assert len(input_vars) > 1, "`add_sequence` requires at least two operands"
        self.variables = input_vars
        out = sum(var.data for var in input_vars)
        return out

    def backward_var(self, grad, index, **kwargs):
        return grad


class MultiplySequence(BroadcastableOp):
    """ Performs f(a, b, ..., z) = a * b * ... * z"""

    def __call__(self, *input_vars):
        self.variables = input_vars
        assert 2 <= len(self.variables)

        out = reduce(lambda x, y: x * y, (var.data for var in input_vars))
        self._iszero = np.any(out == 0)
        return out

    def backward(self, grad, *, graph, **kwargs):
        """ Back-propagates the gradient through all of the operation's inputs. This needs to be updated
            by an operation if that operation takes more than 2 Tensor arguments."""
        if not self._iszero:
            self._product = grad * reduce(
                lambda x, y: x * y, (var.data for n, var in enumerate(self.variables))
            )
        else:
            self._product = None
        super().backward(grad, graph=graph, **kwargs)

    def backward_var(self, grad, index, **kwargs):
        var = self.variables[index]
        if not self._iszero:
            return self._product / var.data
        else:
            return grad * reduce(
                lambda x, y: x * y,
                (var.data for n, var in enumerate(self.variables) if n != index),
            )
